
if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *

class KerasSmartModel(tf.keras.models.Model):
    
    sample_hparams = {
        'model_name': 'KerasSmartModel',
        'l2_reg': 0.0001,
        'l1_reg': None,
        'batch_size': 16,
        'epochs': 2,
        'validation_data': 0.1,
        'early_stopping_patience_epochs': 10,
        'learning_rate': 0.0001,
        'exponential_decay_rate': 0.99,
        'loss_function': 'categorical_crossentropy',
        'loss_function_params': None,
        'metrics':['accuracy'],
        'metrics_params': None,
        'optimizer': 'Adam',
        'optimizer_params': None,
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6,
        'other_callbacks': None,
        'custom_schedule': None
    }
    
    def __init__(self, hparams:dict=None):
        """
        Base class for smart, trainable pytorch modules. All hyperparameters are contained within the `hparams`
        dictionary. Some training-related hyperparameters are common across almost all kinds of PyTorch modules,
        which can be overloaded by the child class. The module includes functions for training, evaluation, and
        prediction. These functions cane be modified or overloaded by any child subclass.

        ### Usage

        `net = KerasSmartModel(hparams)` where `hparams` is dictionary of hyperparameters containing the following:
            - `model_name` (str): Name of the model.
            - `batch_size` (int): Minibatch size, the expected input size of the network.
            - `learning_rate` (float): Initial learning rate of training.
            - `exponential_decay_rate` (float): Exponential decay rate for learning rate, if any.
            - `optimizer` (str): Optimizer. Examples are "Adam", "SGD", "RMSprop", "Adagrad", etc. The name of any Keras optimizer can be used.
                This can also be a custom optimizer class (not instance), in which case `optimizer_params` can be specified for its constructor kwargs.
            - `optimizer_params` (dict): Additional parameters of the optimizer constructor, if any.
            - `epochs` (int): Maximum number of epochs for training.
            - `early_stopping_patience_epochs` (int): Epochs to tolerate unimproved (val) loss, before early stopping (i.e., patience).
            - `l2_reg` (float): L2 regularization parameter. Defaults to None.
            - `l1_reg` (float): L1 regularization parameter. Defaults to None.
            - `loss_function` (str): Loss function. It can be a lower-case name such as "mse", "binary_crossentropy", "categorical_crossentropy", etc.,
                the name of a `tf.keras.losses` class such as `BinaryCrossentropy`, `CategoricalCrossentropy`, `MeanSquaredError', etc., or a valid loss class (not instance).
            - `loss_function_params` (dict): Additional kwargs parameters of the loss function constructor, if any. Ignored if the loss function is a lower-case name string.
            - `metrics` (list): list of metrics for Keras compilation. Each member of the list can be a lower-case metric name such as "mse" or "accuracy", the name of a 
                `tf.keras.metrics` class such as `Accuracy`, `MeanSquaredError`, etc., or a valid metric class (not instance).
            - `metrics_params` (list): (list of) additional kwargs parameters of the metrics constructors, if any. Ignored for every metric that is a lower-case name string.
                If this entry is a single dicitonary rather than a list, it will be broadcast to all metrics in the metrics list.
            - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
            - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
            - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
            - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.
            - `other_callbacks` (list): List of other callbacks to be used during training. Defaults to None.
            - `custom_schedule` (schedule): Custom learning rate schedule inheriting from `tf.keras.optimizers.schedules.LearningRateSchedule`. Defaults to None.

        ### Returns

        Returns a `tf.keras.models.Model` object that can be trained and used accordingly.
        Run `net.summary()` afterwards to see what you have inside the network.
        
        ### Notes:
        
        - `self.batch_input_shape` attribute must be set in the `__init__` method.
        - `self.batch_output_shape` attribute must be set in the `__init__` method.
        - `self._callbacks` is a list of callbacks sent to the training method of Keras. It already includes garbage collection.
        - `self._es` is an EarlyStopping instance of Keras, if specified, otherwise None.
        - `self._chkpt` is the hyperparameter `checkpoint_path` from the input dictionary, if it exists, otherwise None.
        - `self._chk` is a ModelCheckpoint instance sent to the training function of Keras, if specified, otherwise None.
        - `self._es_crit` is an EarlyStopAtCriteria instance that stops the training if 'val_loss' for instance, reaches a certain value.
        - `self.net` **MUST** always exist so that Keras2Cpp can serialize the layers that it understands.
          `self.net` is always used as the model within this module, that contains the network itself. It is typically a Sequential or Functional model built in the `__init__` 
          method.
        """
        super(KerasSmartModel, self).__init__()
        if not hparams: hparams = self.sample_hparams
        self.hparams = hparams
        self._batch_size = int(hparams["batch_size"]) if hparams.get("batch_size") else 32
        self._loss_function = hparams.get("loss_function")
        self._loss_function_params = hparams.get("loss_function_params")
        self._metrics = hparams.get("metrics")
        self._metrics_params = hparams.get("metrics_params")
        self._optimizer = hparams.get("optimizer")
        self._optimizer_params = hparams.get("optimizer_params")
        self._early_stopping_patience_epochs = hparams.get("early_stopping_patience_epochs")
        self._learning_rate = hparams.get("learning_rate")
        self._exponential_decay_rate = hparams.get("exponential_decay_rate")
        self._validation_data = hparams.get("validation_data")
        self._epochs = hparams.get("epochs")
        self._l2_reg = hparams.get("l2_reg") if hparams.get("l2_reg") else None
        self._l1_reg = hparams.get("l1_reg") if hparams.get("l1_reg") else None
        self.history = None
        self.batch_input_shape = (self._batch_size, 1)
        self.batch_output_shape = (self._batch_size, 1)
        self._callbacks = [GarbageCollectionCallback()]
        self._es = tf.keras.callbacks.EarlyStopping(monitor='val_loss' if hparams.get('validation_data') is not None else 'loss', 
                                                    mode="min", patience=self._early_stopping_patience_epochs) if self._early_stopping_patience_epochs else None
        if self._es:
            self._callbacks.append(self._es)
        self._chkpt = hparams.get("checkpoint_path")
        if self._chkpt:
            self._chk = tf.keras.callbacks.ModelCheckpoint(self._chkpt, monitor=('val_loss' if hparams.get('validation_data') is not None else 'loss'), 
                                                           verbose=0, save_best_only=True, mode='min')
        else:
            self._chk = None
        if self._chk:
            self._callbacks.append(self._chk)
        self.early_stopping_monitor = hparams.get("early_stopping_monitor")
        self.early_stopping_mode = hparams.get("early_stopping_mode")
        self.early_stopping_value = hparams.get("early_stopping_value")
        if self.early_stopping_monitor:
            self._es_crit = EarlyStopAtCriteria(monitor=self.early_stopping_monitor, mode=self.early_stopping_mode, value=self.early_stopping_value)
            self._callbacks.append(self._es_crit)
        else:
            self._es_crit = None
        self._other_callbacks = hparams.get("other_callbacks")
        if self._other_callbacks is not None:
            assert isinstance(self._other_callbacks, list), "other_callbacks must be a list of callbacks"
            self._callbacks.extend(self._other_callbacks)
        self._custom_schedule = hparams.get("custom_schedule")
        
    
        # Construct an empty Sequential model
        self.net = tf.keras.models.Sequential()
    
    
    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)
    
    # def build(self):
    #     super().build(input_shape=self.batch_input_shape) 
        
    def summary(self):
        return self.net.summary()
        
    def get_config(self):
        config = super(KerasSmartModel, self).get_config()
        config['hparams'] = self.hparams
        return config
    
    @classmethod
    def from_config(cls, config):
        return cls(config['hparams'])
    
    
    def compile_model(self, num_samples=None):
        """Compiles Keras Smart Model based on its constructor hyperparameters

        Args:
            num_samples (int, optional): Number of samples. Defaults to None.
        """
        compile_keras_model(self.net, _batchsize=self._batch_size, _learnrate=self._learning_rate, _optimizer=self._optimizer, 
                            _loss=self._loss_function, _metrics=self._metrics, _optimizer_params=self._optimizer_params,
                            _loss_params=self._loss_function_params, _metrics_params=self._metrics_params, 
                            _exponential_decay_rate=self._exponential_decay_rate, num_samples=num_samples, custom_schedule=self._custom_schedule)
    
    
    def fit_model(self, x_train, y_train, x_val=None, y_val=None, verbose:int=1, **kwargs):
        """Fit (train) the Keras Smart Model to training data.

        Args:
            x_train (array): Training inputs
            y_train (array): Training target outputs
            x_val (array, optional): Validation inputs. Defaults to None.
            y_val (array, optional): Validation target outputs. Defaults to None.
            verbose (int, optional): Verbosity passed to the Keras Fit function. Defaults to 1.
            
            Other keyword arguments are passed to the Keras fit function.

        Returns:
            History object returned by the Keras fit function
        """
        self.history = fit_keras_model(self.net, x_train, y_train, x_val, y_val, 
            self._batch_size, self._epochs, self._callbacks, verbose, **kwargs)
        return self.history

    
    def __str__(self):
        s = "KerasSmartModel with the following attributes:\n" + str(self.hparams)
        return s
    
    
    def train_model(self, x_train, y_train, x_val=None, y_val=None, verbose:int=1, save_model_to:str=None, save_hparams_to:str=None, 
                    export_to_file:str=None, save_kwargs:dict={}, **kwargs):
        """Train the model according to its hyperparameters.

        ### Args:
            - `x_train` (numpy array): Training inputs
            - `y_train` (numpy array): Training target outputs
            - `x_val` (numpy array): Validation inputs
            - `y_val` (numpy array): Validation target outputs
            - `verbose` (int, optional): Verbosity of training passed to Keras fit function. Defaults to 1.
            - `save_model_to` (str, optional): Save Keras model in path. Defaults to None. Path does not need to exist.
              The path can have no extension, in which case a SavedModel format will be used, or it can have a .h5 extension, in which case a HDF5 format will be used, or it can have a .keras extension, in which case the new Keras format will be used.
            - `save_hparams_to` (str, optional): Save hyperparameters to path. Defaults to None. Path does not need to exist.
            - `export_to_file` (str, optional): Save Keras model in .model file using keras2cpp for later use in C++. Defaults to None.
              Path does not need to exist.
            - `save_kwargs` (dict, optional): Additional kwargs to pass to the `tf.keras.Model.save()` function. Defaults to None.
            
            Other keyword arguments are passed to the Keras `tf.keras.Model.fit()` function.

        ### Returns:
            Nothing. It modifies the "net" attribute of the model, and the history of the training in self.history.
        
        """
        N = x_train.shape[0]
        self.compile_model(num_samples=N)
        _ = self.fit_model(x_train, y_train, x_val, y_val, verbose=verbose, **kwargs)
        if save_model_to:
            save_keras_model(
                model=self.net, 
                history=self.history.history, 
                save_model_to=save_model_to, 
                save_hparams_to=save_hparams_to, 
                hparams=self.hparams,
                **save_kwargs)
        if export_to_file:
            export_keras_model(self.net, export_to_file)


    def plot_history(self, metrics=['loss'], fig_title='model loss', saveto:str=None, close_after_finish:bool=True):
        """Plot the training history of the model after training is done.

        ### Args:
            - `metrics` (list, optional): Metrics list. Defaults to ['loss']. It can also be ['loss','val_loss'] if validation data was provided, or ['accuracy','val_accuracy'].
            - `fig_title` (str, optional): Title of the figure. Defaults to 'model loss'.
            - `saveto` (str, optional): Path to where to save the figure. Defaults to None. Path does not need to exist.
            - `close_after_finish` (bool, optional): Clsoe the figure after saving it with `plt.close()`. Defaults to True. Only applicable if `saveto` is provided.
        """
        plot_keras_model_history(self.history.history, metrics, fig_title, saveto, close_after_finish)
      
        
    def evaluate(self, *args, **kwargs):
        """Evaluate model performance on test data.

        Returns:
            The same thing that Keras evaluate returns.
        """
        return self.net.evaluate(*args, **kwargs)
    
    
    def predict(self, *args, **kwargs):
        """Predict the outputs given inputs

        Returns:
            Teh same thing that Keras predict returns.
        """
        return self.net.predict(*args, **kwargs)
    
    
    def make_regularizer(self):
        if self._l1_reg is not None and self._l1_reg > 0 and (self._l2_reg is None or self._l2_reg == 0): # Only do L1 regularization
            return tf.keras.regularizers.L1(self._l1_reg)
        elif self._l2_reg is not None and self._l2_reg > 0 and (self._l1_reg is None or self._l1_reg == 0): # Only do L2 regularization
            return tf.keras.regularizers.L2(self._l2_reg)
        elif self._l1_reg is not None and self._l1_reg > 0 and self._l2_reg is not None and self._l2_reg > 0: # Do both L1 and L2 regularization
            return tf.keras.regularizers.L1L2(self._l1_reg, self._l2_reg)
        else:
            return None 
    
    
    def save_model(self, save_model_to:str=None, save_hparams_to:str=None, export_to_file:str=None, **kwargs):
        """Save the model to file.

        ### Args:
            - `save_model_to` (str, optional): Save Keras model in path. Defaults to None. Path does not need to exist.
              The path can have no extension, in which case a SavedModel format will be used, or it can have a .h5 extension, in which case a HDF5 format will be used, or it can have a .keras extension, in which case the new Keras format will be used.
            - `save_hparams_to` (str, optional): Save hyperparameters to path. Defaults to None. Path does not need to exist.
            - `export_to_file` (str, optional): Save Keras model in .model file using keras2cpp for later use in C++. Defaults to None.
              Path does not need to exist.
            - **kwargs: Additional kwargs to pass to the `tf.keras.Model.save()` function. Defaults to None.
        """
        save_keras_model(
            model=self.net, 
            history=None, 
            save_model_to=save_model_to, 
            save_hparams_to=save_hparams_to, 
            hparams=self.hparams,
            **kwargs)
        if export_to_file:
            export_keras_model(self.net, export_to_file)
    
    
    def export_model(self, export_to_file:str):
        """Export the model to file.

        ### Args:
            - `export_to_file` (str): Save Keras model in .model file using keras2cpp for later use in C++. Path does not need to exist.
        """
        export_keras_model(self.net, export_to_file)
        





# Global Constants
SEED = 42

# General-Purpose Libraries
if __name__ == "eznet_keras.utils":
    from .keras2cpp import export_model
else:
    from keras2cpp import export_model
import os
import warnings
from pathlib import Path
import math
import json
import random
random.seed(SEED)
import numpy as np
np.random.seed(SEED)
from timeit import default_timer as timer
from datetime import datetime
from matplotlib import pyplot as plt

# Tensorflow Libraries
import tensorflow as tf
tf.random.set_seed(SEED)
import gc

# Reset seeds, just in case they were modified during imports
random.seed(SEED)
np.random.seed(SEED)
tf.random.set_seed(SEED)
GLOROTUNIFORM = tf.keras.initializers.GlorotUniform(seed=SEED)
ORTHOGONAL = tf.keras.initializers.Orthogonal(seed=SEED)


########################################################################################################################
# Global variables, functions, and classes
########################################################################################################################

optdict_keras = {'adam':tf.keras.optimizers.Adam, 'sgd':tf.keras.optimizers.SGD, 'rmsprop':tf.keras.optimizers.RMSprop, 'adagrad':tf.keras.optimizers.Adagrad}
actdict_keras = {
    'relu':tf.keras.activations.relu, 'leakyrelu':tf.keras.layers.LeakyReLU, 
    'sigmoid':tf.keras.activations.sigmoid, 'tanh':tf.keras.activations.tanh, 'softmax':tf.keras.activations.softmax,
    'softplus':tf.keras.activations.softplus, 'softsign':tf.keras.activations.softsign,
    'elu':tf.keras.activations.elu, 'selu':tf.keras.activations.selu}
rnndict_keras = {'LSTM':tf.keras.layers.LSTM, 'GRU':tf.keras.layers.GRU, 'SimpleRNN':tf.keras.layers.SimpleRNN}
convdict_keras = {"conv1d":tf.keras.layers.Conv1D, "conv2d":tf.keras.layers.Conv2D, "conv3d":tf.keras.layers.Conv3D}


def make_path(path:str):
    Path.mkdir(Path(path).parent, parents=True, exist_ok=True)
    return path
    

def plot_keras_model_history(history:dict, metrics:list=['loss'], fig_title:str='model loss', saveto:str=None, close_after_finish:bool=True):
    plt.figure(figsize=(7, 5))
    plt.grid(True)
    plt.plot(history[metrics[0]], label='training')
    if len(metrics) > 1:
        plt.plot(history[metrics[1]], label='validation')
    plt.title(fig_title)
    plt.ylabel(metrics[0])
    plt.xlabel('epoch')
    plt.legend(loc='upper right')
    if saveto:
        plt.savefig(make_path(saveto), dpi=600)
        if close_after_finish:
            plt.close()
        



def compile_keras_model(model, _batchsize:int, _learnrate:float, _optimizer:str, _loss:str, _metrics:list, 
                          _optimizer_params:dict=None, _loss_params:dict=None, _metrics_params:list=None, _exponential_decay_rate:float=None, num_samples:int=None, custom_schedule=None):
    if custom_schedule:
        lr = custom_schedule
    elif _exponential_decay_rate:
        itersPerEpoch = (num_samples//_batchsize) if num_samples else 1
        sch = tf.keras.optimizers.schedules.ExponentialDecay(initial_learning_rate=_learnrate, 
        decay_steps=itersPerEpoch, decay_rate=_exponential_decay_rate)
        lr = sch
    else:
        lr = _learnrate
    
    _opt_module = getattr(tf.keras.optimizers, _optimizer) if isinstance(_optimizer, str) else _optimizer
    
    if _optimizer_params:
        optparam = _optimizer_params
        # opt = optdict_keras[_optimizer](learning_rate=lr, **optparam)
        opt = _opt_module(learning_rate=lr, **optparam)
    else:
        # opt = optdict_keras[_optimizer](learning_rate=lr)
        opt = _opt_module(learning_rate=lr)
    
    if isinstance(_loss, str):
        if _loss.lower()==_loss: # Put it as is
            _lossfunc = _loss
        else: # It is the name of a Keras losses class
            _loss_module = getattr(tf.keras.losses, _loss)
            if _loss_params:
                _lossfunc = _loss_module(**_loss_params)
            else:
                _lossfunc = _loss_module()
    else: # It is a custom loss function class
        if _loss_params:
            _lossfunc = _loss(**_loss_params)
        else:
            _lossfunc = _loss()
        
    
    assert isinstance(_metrics, list), "Metrics must be a list of strings or a list of metrics"
    if _metrics_params and isinstance(_metrics_params, dict):
        _metrics_params = [_metrics_params]
    if _metrics_params:
        assert len(_metrics_params)==1 or len(_metrics_params)==len(_metrics), \
            "If metrics_params is specified, it must be a list of length 1 or the same length as metrics."
    if _metrics_params:
        _metrics_params_vec = _metrics_params if len(_metrics_params)==len(_metrics) else [_metrics_params[0]]*len(_metrics)
            
    _metrics_list = []
    for i,metric in enumerate(_metrics):
        if isinstance(metric, str):
            if metric.lower()==metric: # Put it as is
                metr = metric
            else: # It is the name of a Keras metrics class
                _metr_module = getattr(tf.keras.metrics, metric)
                if _metrics_params_vec[i] is not None:
                    metr = _metr_module(**_metrics_params_vec[i])
                else:
                    metr = _metr_module()
        else: # It is a custom metric class
            if _metrics_params_vec[i] is not None:
                metr = metric(**_metrics_params_vec[i])
            else:
                metr = metric()
        _metrics_list.append(metr)
    
    
    model.compile(optimizer=opt, loss=_lossfunc, metrics=_metrics_list)
    



def fit_keras_model(model, x_train, y_train, x_val=None, y_val=None, 
    _batchsize:int=None, _epochs:int=1, _callbacks:list=None, verbose:int=1, **kwargs):
    while True:
        try:
            history = model.fit(x_train, y_train, batch_size=_batchsize, epochs=_epochs, 
                validation_data=((x_val, y_val) if x_val is not None and y_val is not None else None), verbose=verbose, 
                callbacks=_callbacks, **kwargs)
            break
        except Exception as e:
            print(e)
            print(("\nTraining failed with batchsize={}. "+\
                "Trying again with a lower batchsize...").format(_batchsize))
            _batchsize = _batchsize // 2
            if _batchsize < 2:
                raise ValueError("Batchsize too small. Training failed.")
    return history


def save_keras_model(model, save_model_to:str, save_hparams_to:str=None, history:dict=None, hparams:dict=None, **kwargs):
    try:
        model.save(make_path(save_model_to), **kwargs)
        if history is not None:
            assert hparams is not None, "If training history is to be saved along with the hyperparameters, then `hparams` must be provided."
            for key in history:
                hparams[key] = history[key]
        if hparams is not None:
            assert save_hparams_to is not None, "`save_hparams_to` must be provided to where the hyperparameters are to be saved."
            if not save_hparams_to.endswith(".json"):
                warnings.warn("save_hparams_to does not end with '.json'. A json extension is highly recommended for hyperparameters.", UserWarning)
            jstr = json.dumps(hparams, indent=4)
            with open(make_path(save_hparams_to), "w") as f:
                f.write(jstr)
    except Exception as e:
        print("Cannot serialize Keras model.")
        raise e
        
        
def export_keras_model(model, path:str):
    try:
        export_model(model, make_path(path))
        print("Model exported successfully.")
    except Exception as e1:
        print(e1)
        print("Cannot export Keras model using keras2cpp on the fly. Will try sequentializing the model layers...")
        try:
            net = tf.keras.models.Sequential(model.layers)
            export_model(net, make_path(path))
            print("Model exported successfully.")
        except Exception as e2:
            print("Cannot export model using Keras2Cpp.")
            raise e2
    

def test_keras_model_class(model_class, hparams:dict=None, save_and_export:bool=True):
    print("Constructing model...\n")
    model = model_class(hparams)
    print("Summary of model:")
    print(model.summary())
    print("\nGenerating random dataset...\n")
    (x_train, y_train) = generate_sample_batch(model)
    (x_val, y_val) = generate_sample_batch(model)
    print("Trying forward pass on training data: ")
    y = model(x_train)
    print("\nOutput shape: ", y.shape)
    print("\nTraining model...\n")
    model_name = model.hparams["model_name"]
    model.train_model(x_train, y_train, x_val, y_val, 
                verbose=1, 
                save_model_to=(f"test_{model_name}" if save_and_export else None),
                save_hparams_to=(f"test_{model_name}_hparams.json" if save_and_export else None), 
                export_to_file=(f"test_{model_name}.model" if save_and_export else None))
    print("\nEvaluating model...\n")
    model.evaluate(x_val, y_val, verbose=1)
    print("Done.")
    

# Perform garbage collection
import gc
class GarbageCollectionCallback(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        gc.collect()
        
        
# Deploy early stopping when performance reaches good values
class EarlyStopAtCriteria(tf.keras.callbacks.Callback):
    def __init__(self, monitor='val_loss', mode='min', value=0.001):
        super(EarlyStopAtCriteria, self).__init__()
        self.monitor = monitor
        self.value = value
        self.mode = mode
    def on_epoch_end(self, epoch, logs=None):
        if self.mode == 'min':
            if logs.get(self.monitor) <= self.value:
                print("Early stopping performance criteria has been reached. Stopping training.")
                self.model.stop_training = True
        else:
            if logs.get(self.monitor) >= self.value:
                print("Early stopping performance criteria has been reached. Stopping training.")
                self.model.stop_training = True
        

# Sampling layer used for variational autoencoders
class Sampling(tf.keras.layers.Layer):
    """Uses (z_mean, z_log_var) to sample z, the vector encoding a digit."""

    def call(self, inputs):
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


def autoname(name):
    """
    Genereate a unique name for a file, based on the current time and the given name.
    Gets the `name` as a string and adds the time stamp to the end of it before returning it.
    """
    return name + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S")



def calc_image_size(size_in:int, kernel_size:int, padding:int, stride:int, dilation:int):
    """Calculate image size after convolution or pooling.
    ### Args:
        - `size_in` (int|list): (list of) image input dimension(s).
        - `kernel_size` (int|list): (list of) kernel (or pool) size
        - `padding` (int|list): (list of) padding sizes, or string such as 'valid' and 'same'.
        - `stride` (int|list): (list of) strides.
        - `dilation` (int|list): (list of) dilation rates.
    ### Returns:
        int or list: Output image dimensions
    """
    if padding == 'same':
        return size_in
    else:
        if padding == 'valid':
            padding = 0
        if isinstance(size_in, (list, tuple)):
            if isinstance(padding, int): padding = [padding]*len(size_in)
            if isinstance(kernel_size, int): kernel_size = [kernel_size]*len(size_in)
            if isinstance(stride, int): stride = [stride]*len(size_in)
            if isinstance(dilation, int): dilation = [dilation]*len(size_in)
            return [math.floor((size_in[i] + 2*padding[i] - dilation[i]*(kernel_size[i]-1) - 1)/stride[i] + 1) for \
                i in range(len(size_in))]
        else:
            assert isinstance(size_in, int), "size_in must be an integer or a list/tuple of integers."
            assert isinstance(padding, int), "padding must be an integer or a list/tuple of integers."
            assert isinstance(kernel_size, int), "kernel_size must be an integer or a list/tuple of integers."
            assert isinstance(stride, int), "stride must be an integer or a list/tuple of integers."
            assert isinstance(dilation, int), "dilation must be an integer or a list/tuple of integers."
            return math.floor((size_in + 2*padding - dilation*(kernel_size-1) - 1)/stride + 1)


def generate_geometric_array(init, count, direction, powers_of_two=True):
    """Generate array filled with incrementally doubling/halving values, optionally with powers of two.

    ### Args:
    
        - `init` (int): The first value to begin.
        - `count` (int): Number of elements to generate.
        - `direction` (str): Direction of the array. Can be either 'up' or 'down', i.e. increasing or decreasing.
        - `powers_of_two` (bool, optional): Generate numbers that are powers of two. Defaults to True.

    ### Returns:
    
        list: List containing elements
    """
    lst = []
    old = int(2**math.ceil(math.log2(init))) if powers_of_two else init
    new = old
    for _ in range(count):
        lst.append(new)
        old = new
        new = (old * 2) if direction == 'up' else (old // 2)
    return lst


def generate_array_for_hparam(
    hparam, count_if_not_list:int, 
    hparam_name:str='parameter', count_if_not_list_name:str='its count',
    check_auto:bool=False, init_for_auto:int=2, powers_of_two_if_auto:bool=True,
    direction_if_auto:str=None):
    """Generate array for a hyperparameter, regardless of if it is a list or not. This function is for use in APIs
    that generate models with hyperparameters as inputs, which can be lists, a single item, or "auto".
    Examples include width of a neural network's hidden layers, channels of conv layers, etc.
    For these hyperparameters, the user is typically free to specify an array-like, a single item to be repeated,
    or "auto" for automatic calculation of the parameter.
    This function is meant to be used in the body of the code of class constructors and other functions in the API.

    ### Args:
    
        - `hparam` (var): A specific hyperparameter, e.g., input by user to your network's constructor.
        - `count_if_not_list` (int): Number of elements to generate if `hparam` is not an array-like.
        - `hparam_name` (str, optional): Name of the hyperparameter. Defaults to 'parameter'.
        - `count_if_not_list_name` (str, optional): Name of the "count" that must be provided. Defaults to 'its count'.
        - `check_auto` (bool, optional): Check for the "auto" case. Defaults to False.
        - `init_for_auto` (int, optional): Initial value in case of "auto". Defaults to 2.
        - `powers_of_two_if_auto` (bool, optional): Generate powers of two in case of "auto". Defaults to True.
        - `direction_if_auto` (str, optional): Direction of geometric increment in case of "auto". Defaults to None.
           This can be "up" or "down". If check_for_auto is True, then this argument must be specified.

    ### Returns:
    
        list: List containing elements
    """
    assert count_if_not_list is not None, \
        "Since %s may not be a list/tuple, %s must always be specified."%(hparam_name, count_if_not_list_name)
    if isinstance(hparam, (list,tuple)) and len(hparam) == count_if_not_list:
        lst = hparam
    elif hparam == "auto" and check_auto:
        assert init_for_auto is not None, \
            "If %s is 'auto', then %s must be specified."%(hparam_name, "init_for_auto")
        assert direction_if_auto in ['up','down'], \
            "If %s is 'auto', then %s must be specified as 'up' or 'down'."%(hparam_name, "direction_if_auto")
        lst = generate_geometric_array(init_for_auto, count_if_not_list, direction_if_auto, powers_of_two_if_auto)
    else:
        lst = [hparam]*count_if_not_list
    return lst




def generate_sample_batch(model):
    x = np.random.rand(*model.batch_input_shape).astype(np.float32)
    y = np.random.rand(*model.batch_output_shape).astype(np.float32)
    if model.hparams['loss_function'] in ['sparse_categorical_crossentropy']:
        y = np.argmax(y, axis=1)
    return (x,y)
    


if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *
import warnings

class Conv_Block(tf.keras.layers.Layer):
    def __init__(self, input_shape:list, out_channels:int=None, conv_dim:int=1, input_image:list=[30], conv_kernel_size=3, conv_padding='valid', conv_stride=1, conv_dilation=1, 
                 conv_params:dict=None, conv_activation:str='relu', conv_activation_params:dict=None, norm_layer_position:str=None, norm_layer_type:str=None, 
                 norm_layer_params:dict=None, pool_type:str=None, pool_kernel_size=2, pool_padding:str='valid', pool_stride=1, pool_params:dict=None, 
                 dropout:float=None, spatial_dropout:float=None, min_image_dim:int=1, kernel_regularizer:tf.keras.regularizers.Regularizer=None):
        """Convolutional block, used as a single building block for all kinds of convolutional neural networks.
        It contains one convolution layer, followed optionally by a normalization layer, an activation layer, a pooling layer and a dropout layer.
        The convolution layer is mandatory, but the other ones are optional. The convolution layer can be 1D, 2D or 3D. 
        The normalization layer can be the name of any such layer defined in Keras, e.g. `BatchNormalization`, or a custom `tf.keras.Layer` class (not instance).
        The pooling type can be `Avg`, `Max`, `GlobalAvg`, `GlobalMax`, or a custom `tf.keras.Layer` class (not instance).
        The activation can be the name of an activation function (e.g. 'relu', 'sigmoid', etc.), or the name of a Keras activation layer (e.g. 'ReLU', 'LeakyReLU', etc.), 
        or a custom Keras Layer class (not instance).
        The dropout consists of a spatial dropout layer followed by a normal dropout layer. It is, however, not common to use both types of dropout in a single block. 
        The dimension of any dropout layer will match the dimension of the convolution layer. For `Conv1D` for instance, `SpatialDropout1D` will be used, if desired.
        
        This module is meant to be used as a building block for larger modules.
        
        **NOTE**: 
        Activation functions come from `tf.keras.activations` and need to be wrapped in a `tf.keras.layers.Activation` layer, and do not accept any kwargs. Activation layers
        come from `tf.keras.layers`, can accept kwargs in their constructors, and do not need to be wrapped in an `Activation` layer. Activation functions all have lower-case
        names, but activation layers are classes and have every-word-capitalized names. We will use this as a clue to recognize what the user wants, and perform accordingly.
        Most activation layers also have correpsonding activation functions (like `tf.keras.activations.relu` function and `tf.keras.layers.ReLU` layer). 
        However, some activations are only available as functions (such as `tf.keras.activations.sigmoid`), and some are only available as layers 
        (such as `tf.keras.layers.LeakyReLU`). Choose accordingly.

        ### Args:
        
        - `input_shape` (list|tuple): Input shape.
        - `out_channels` (int, optional): Number of convolution filters. Defaults to the input channels size.
        - `conv_dim` (int, optional): Dimension of the convolution. Defaults to 1. 1 means Conv1d, 2 means Conv2d etc.
        - `input_image` (list, optional): Size of the input image. Defaults to [30]. This must be a list/tuple of integers, with legth equal to `conv_dim`.
        - `conv_kernel_size` (int, optional): Convolution kernel size. Defaults to 3. It is strongly recommended to provide a list of integers, with length equal to `conv_dim`.
        - `conv_padding` (str, optional): Convolution padding. Defaults to 'same'. Arrays are recommended over integers.
        - `conv_stride` (int, optional): Convolution stride. Defaults to 1.
        - `conv_dilation` (int, optional): Convolution dilation. Defaults to 1.
        - `conv_params` (dict, optional): Additional dictionary of kwargs for Conv?d module, to add or overwrite other parameters. Defaults to None.
          No matter what parameters are set by this class by default for Con?D layer, providing this dictionary will overwrite all of them.
        - `conv_activation` (str, optional): Name of activation layer class ("ReLU", "LeakyReLU", "Softmax", etc.), or activation function ("relu", "sigmoid", etc.),
          or custom Keras Layer class (not instance). Defaults to 'ReLU'.
        - `conv_activation_params` (dict, optional): kwargs dictionary for constructor of activation layer, if it is a layer or the name of a defined layer. 
          Otherwise, if activation is the name of a function such as 'sigmoid', this entry will be ignored. Defaults to None. For `LeakyReLU`, `alpha` is the slope parameter
          for the negative part.
        - `norm_layer_position` (str, optional): Position of the normalization layer relative to activation. Defaults to None. It should be 'before' or 'after' or None.
        - `norm_layer_type` (str, optional): Type of the normalization layer. Defaults to None. Examples: 'BatchNormalization', 'LayerNormalization', etc.
          This can also be a custom-made Keras Layer class (not instance).
        - `norm_layer_params` (dict, optional): kwargs dictionary for normalization layer constructor. Defaults to None.
        - `pool_type` (str, optional): Type of pooling layer, if any. Defaults to None. For example, 'Max', 'Avg', 'GlobalMax', 'GlobalAvg' etc.
          This can also be a custom-made Keras Layer class (not instance).
        - `pool_kernel_size` (int, optional): Pooling kernel size. Defaults to 2. Arrays are recommended over integers.
        - `pool_padding` (int, optional): Padding for pooling layer. Defaults to 0. 'same' is NOT an option here.
        - `pool_stride` (int, optional): Pooling stride. Defaults to 1.
        - `pool_params` (dict, optional): kwargs dictionary for pooling layer constructor to add to or overwrite its arguments. Defaults to None.
        - `dropout` (float, optional): Dropout rate, if any. Defaults to None.
        - `spatial_dropout` (float, optional): Spatial dropout rate, if any. Defaults to None. For Conv?D, SpatialDropout?D is used.
        - `min_image_dim` (int, optional): Minimum image dimension. Defaults to 1. This is used for preventing the image dimension from becoming too small. 
            It can automatically adjust padding and stride for convolution and pooling layers to keep the image dimensions larger than this argument.
        - `kernel_regularizer` (regularizer, optinal): Regularizer to be used for the kernel weights in the layers.
        
        ### Returns:
        A `tf.keras.layers.Layer` instance representing a single convolutional block.
        
        ### Attributes:
        `self.output_image` (list): Size of the output image.
        `self.net` (tf.keras.models.Sequential): The actual network, a `tf.keras.models.Sequential` instance.
        """
        super(Conv_Block, self).__init__()
        self.net = tf.keras.models.Sequential()
        ret = add_conv_block(self.net, out_channels, input_shape, conv_dim, input_image, conv_kernel_size, conv_padding, conv_stride, conv_dilation, conv_params,
                             conv_activation, conv_activation_params, norm_layer_position, norm_layer_type, norm_layer_params, 
                             pool_type, pool_kernel_size, pool_padding, pool_stride, pool_params, dropout, spatial_dropout, min_image_dim, kernel_regularizer)
        self.net = ret['model']
        self._input_shape = input_shape
        self.output_image = ret['output_image']
        self._out_channels = ret['out_channels']
        self._norm_layer_position = ret['norm_layer_position']
        self._norm_layer_type = ret['norm_layer_type']
        self._min_image_dim = ret['min_image_dim']
        self._conv_padding = ret['conv_padding']
        self._conv_stride = ret['conv_stride']
        self._pool_padding = ret['pool_padding']
        self._pool_stride = ret['pool_stride']
        self._conv_module = ret['conv_module']
        self._conv_activation_module = ret['conv_activation_module']
        self._pool_module = ret['pool_module']
        self._dropout_module = ret['dropout_module']
        self._spatial_dropout_module = ret['spatial_dropout_module']
        self._norm_layer_module = ret['norm_layer_module']
        
        # Store parameters as class attributes
        self._conv_kernel_size = conv_kernel_size
        self._conv_dilation = conv_dilation
        self._conv_params = conv_params
        self._conv_activation = conv_activation
        self._conv_activation_params = conv_activation_params
        self._norm_layer_params = norm_layer_params
        self._pool_type = pool_type
        self._pool_kernel_size = pool_kernel_size
        self._pool_params = pool_params
        self._dropout = dropout if dropout else None
        self._spatial_dropout = spatial_dropout if spatial_dropout else None
        self._kernel_regularizer = kernel_regularizer

    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)
    
    def get_config(self):
        config = super(Conv_Block, self).get_config()
        hparams = {
            "input_shape":self._input_shape,
            "out_channels":self._out_channels,
            "conv_dim":self._conv_dim,
            "input_image":self._input_image,
            "conv_kernel_size":self._conv_kernel_size,
            "conv_padding":self._conv_padding,
            "conv_stride":self._conv_stride,
            "conv_dilation":self._conv_dilation,
            "conv_params":self._conv_params,
            "conv_activation":self._conv_activation,
            "conv_activation_params":self._conv_activation_params,
            "norm_layer_position":self._norm_layer_position,
            "norm_layer_type":self._norm_layer_type,
            "norm_layer_params":self._norm_layer_params,
            "pool_type":self._pool_type,
            "pool_kernel_size":self._pool_kernel_size,
            "pool_padding":self._pool_padding,
            "pool_stride":self._pool_stride,
            "pool_params":self._pool_params,
            "dropout":self._dropout,
            "spatial_dropout":self._spatial_dropout,
            "min_image_dim":self._min_image_dim,
            "kernel_regularizer":self._kernel_regularizer
        }
        config['hparams'] = hparams
        return config
    
    @classmethod
    def from_config(cls, config):
        return cls(**config['hparams'])
    
    def summary(self):
        return self.net.summary()



def add_conv_block(model:tf.keras.models.Sequential, out_channels:int=None, input_shape:list=None, conv_dim:int=1, 
    input_image:list=[30], conv_kernel_size=3, conv_padding='valid',
    conv_stride=1, conv_dilation=1, conv_params:dict=None, conv_activation:str='ReLU', conv_activation_params:dict=None, norm_layer_position:str=None, norm_layer_type:str=None, 
    norm_layer_params:dict=None, pool_type:str=None, pool_kernel_size=2, pool_padding:str='valid', pool_stride=1, pool_params:dict=None, dropout:float=None, 
    spatial_dropout:float=None, min_image_dim:int=1, kernel_regularizer:tf.keras.regularizers.Regularizer=None):
    """Add a Convolutional block, containing one convolution layer, followed optionally by a normalization layer, an activation layer, a pooling layer and a dropout layer,
    to the end of an existing `tf.keras.models.Sequential` instance.
    The convolution layer is mandatory, but the other ones are optional. The convolution layer can be 1D, 2D or 3D. 
    The normalization layer can be the name of any layer defined such as `BatchNormalization`, or a custom Keras Layer class (not instance). 
    The activation layer can be the name of an activation function ('sigmoid', 'relu', 'tanh', etc.), the name of an activation layer ('ReLU', 'LeakyReLU', 'Softmax', etc.), 
    or a custom Keras Layer class (not instance). 
    The pooling layer can be `Avg`, `Max`, `GlobalAvg` or `GlobalMax`. Also, it can be a custom Keras Layer class (not instance), in which case all pooling-related arguments
    of this function will be ignored, and you'll have to pass pooling layer constructor arguments via `pool_params`. Also, in this case the image size calculations are bypassed.
    The dropout consists of a spatial dropout followed by a normal dropout. It is not common, however, to use both kinds of dropout in a single CNN block.
    The dimension of any dropout layer will match the dimension of the convolution layer. For `Conv1D` for instance, `SpatialDropout1D` will be used, if desired.

    ### Args:
    
    - `input_shape` (list|tuple): Input shape.
    - `out_channels` (int, optional): Number of convolution filters. Defaults to the input channels size.
    - `conv_dim` (int, optional): Dimension of the convolution. Defaults to 1. 1 means Conv1d, 2 means Conv2d etc.
    - `input_image` (list, optional): Size of the input image. Defaults to [30]. This must be a list/tuple of integers, with legth equal to `conv_dim`.
    - `conv_kernel_size` (int, optional): Convolution kernel size. Defaults to 3. It is strongly recommended to provide a list of integers, with length equal to `conv_dim`.
    - `conv_padding` (str, optional): Convolution padding. Defaults to 'same'. Arrays are preferred over integers.
    - `conv_stride` (int, optional): Convolution stride. Defaults to 1.
    - `conv_dilation` (int, optional): Convolution dilation. Defaults to 1.
    - `conv_params` (dict, optional): Additional dictionary of kwargs for constructor of Conv?d module. Defaults to None.
      If provided, this dictionary will not only add to, but also overwrite any existing arguments passed to the Conv?D constructor.
    - `conv_activation` (str, optional): String representing activation function ('sigmoid', 'relu', 'softmax', 'tanh', etc.) or layer ('ReLU', 'LeakyReLU', 'Softmax', etc.).
      Instead of a string, one can also provide a custom Keras Layer class (not instance). Defaults to 'ReLU'.
    - `conv_activation_params` (dict, optional): kwargs dictionary for activation layer constructor, if it is a layer. Defaults to None. 
      If the passed argument is a lower-case function name, this argument will be ignored. By the way, for `LeakyReLU`, `alpha` is the slope parameter in negative part.
    - `norm_layer_position` (str, optional): Position of the normalization layer relative to activation. Defaults to None. It should be 'before' or 'after' or None.
    - `norm_layer_type` (str, optional): Type of the normalization layer. Defaults to None. Examples: 'BatchNormalization', 'LayerNormalization', etc.
      Instead of a string, a custom Keras Layer class (not instance) can also be provided.
    - `norm_layer_params` (dict, optional): kwargs dictionary for normalization layer constructor. Defaults to None.
    - `pool_type` (str, optional): Type of pooling layer, if any. Defaults to None. For example, 'Max', 'Avg', 'GlobalMax', 'GlobalAvg' etc.
      Instead of a string, a custom Keras Layer class (not instance) can also be provided.
    - `pool_kernel_size` (int, optional): Pooling kernel size. Defaults to 2. Arrays are recommended over integers.
    - `pool_padding` (int, optional): Padding for pooling layer. Defaults to 0. 'same' is NOT an option here.
    - `pool_stride` (int, optional): Pooling stride. Defaults to 1.
    - `pool_params` (dict, optional): kwargs dictionary for pooling layer constructor, to add to or overwrite existing parameters. Defaults to None.
    - `dropout` (float, optional): Dropout rate, if any. Defaults to None.
    - `spatial_dropout` (float, optional): Spatial dropout rate, if any. Defaults to None. For Conv?D, SpatialDropout?D will be used.
    - `min_image_dim` (int, optional): Minimum image dimension. Defaults to 1. This is used for preventing the image dimension from becoming too small. 
        It can automatically adjust padding and stride for convolution and pooling layers to keep the image dimensions larger than this argument.
    - `kernel_regularizer` (regularizer, optional): Keras Regularizer to be used for the kernel weights in the layers.
    
    **NOTE**: 
        Activation functions come from `tf.keras.activations` and need to be wrapped in a `tf.keras.layers.Activation` layer, and do not accept any kwargs. Activation layers
        come from `tf.keras.layers`, can accept kwargs in their constructors, and do not need to be wrapped in an `Activation` layer. Activation functions all have lower-case
        names, but activation layers are classes and have every-word-capitalized names. We will use this as a clue to recognize what the user wants, and perform accordingly.
        Most activation layers also have correpsonding activation functions (like `tf.keras.activations.relu` function and `tf.keras.layers.ReLU` layer). 
        However, some activations are only available as functions (such as `tf.keras.activations.sigmoid`), and some are only available as layers 
        (such as `tf.keras.layers.LeakyReLU`). Choose accordingly.
    
    ### Returns:
    A dictionary with the following fields:
    ```python
    {
        'model'
        'output_image'
        'out_channels'
        'norm_layer_position'
        'norm_layer_type'
        'min_image_dim'
        'conv_padding'
        'conv_stride'
        'pool_padding'
        'pool_stride'
        'conv_module'
        'conv_activation_module'
        'pool_module'
        'dropout_module'
        'spatial_dropout_module'
        'norm_layer_module'
    }
    ```
    """
    # Output channels check
    if not (isinstance(out_channels,int) and out_channels > 0):
        warnings.warn("Invalid value out_channels={}. Using value {} equal to input_shape[-1].".format(out_channels,input_shape[-1]), UserWarning)
        out_channels = input_shape[-1]
    # Convolution dimension check    
    assert isinstance(conv_dim, int) and conv_dim in [1,2,3], "`conv_dim` must be an integer among [1,2,3], not {} which has type {}.".format(conv_dim, str(type(conv_dim)))
    # Determine convolution module used
    _conv_module = convdict_keras["conv{}d".format(conv_dim)]
    # Input image size check
    assert isinstance(input_image, (list,tuple)) and len(input_image)==conv_dim, \
        "`input_image` must be a list or tuple of length equal to `conv_dim`, not {} which has type {}.".format(input_image, str(type(input_image)))
    # Check activation function and module
    if isinstance(conv_activation, str):
        if conv_activation.lower() == conv_activation: # activation function such as `sigmoid`
            _conv_activation_module = getattr(tf.keras.activations, conv_activation)
        else: # activation layer such as `ReLU`
            _conv_activation_module = getattr(tf.keras.layers, conv_activation)
    else:
        _conv_activation_module = conv_activation
    # Check position, type and parameters of normalization layer
    if not norm_layer_position in ['before', 'after', None]:
        warnings.warn(("Invalid value {} for `norm_layer_position`: It can only be 'before' (before activation), 'after' (after activation) or None. "+
                        "Using default value of None. There will be no normalization layer.").format(norm_layer_position), UserWarning)
        norm_layer_position = None
    if norm_layer_position is None: norm_layer_type = None
    if norm_layer_type is None: norm_layer_position = None
    if isinstance(norm_layer_type, str):
        _norm_layer_module = getattr(tf.keras.layers, norm_layer_type)
    elif norm_layer_type is not None:
        _norm_layer_module = norm_layer_type
    else:
        _norm_layer_module = None
    # Check pooling layer type, module and parameters
    if isinstance(pool_type, str):
        _pool_module = getattr(tf.keras.layers, "{}Pool{}D".format(pool_type, conv_dim))
    elif pool_type is not None:
        _pool_module = pool_type
    else:
        _pool_module = None
    # Check Dropout parameters
    _spatial_dropout_module = getattr(tf.keras.layers, 'SpatialDropout{}D'.format(conv_dim)) if spatial_dropout else None
    _dropout_module = tf.keras.layers.Dropout if dropout else None
    # Store minimum desired image size
    min_image_dim = min_image_dim if min_image_dim>0 else 1
    img_size = input_image
    # -----------------------------------------------------------------------------        
    # Check if output image size is smaller than min_image_dim, and adjust parameters if necessary
    # print("Running calc_size for input image size {}".format(img_size))
    temp_img_size = calc_image_size(img_size, kernel_size=conv_kernel_size, stride=conv_stride, padding=conv_padding, dilation=conv_dilation)
    if min(temp_img_size) < min_image_dim:
        warnings.warn(
            "Output image ({}) of the convolution operation is smaller in one or more dimensions than min_image_dim={} for add_conv_block(). ".format(temp_img_size,min_image_dim)+ 
            "Using padding='same' and stride=1 instead of padding={} and stride={}".format(conv_padding, conv_stride), UserWarning)
        conv_padding = 'same'
        conv_stride = 1
    # Construct convolutional layer
    _kwargs = {'filters':out_channels, 'kernel_size':conv_kernel_size, 'strides':conv_stride, 'padding':conv_padding, 'dilation_rate':conv_dilation, 
               'kernel_regularizer':kernel_regularizer}
    if input_shape: _kwargs.update({'input_shape':input_shape})
    if conv_params: _kwargs.update(conv_params)
    model.add(_conv_module(**_kwargs))
    # Calculate output image size
    img_size = calc_image_size(img_size, kernel_size=conv_kernel_size, stride=conv_stride, padding=conv_padding, dilation=conv_dilation)
    # ---------------------------------------------------------------------------
    # Construct normalization layer, if it should be here.
    if norm_layer_position=='before':
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
    # Construct activation layer
    if conv_activation:
        if isinstance(conv_activation, str) and conv_activation.lower() == conv_activation:
            model.add(tf.keras.layers.Activation(_conv_activation_module))
        elif conv_activation_params: 
            model.add(_conv_activation_module(**conv_activation_params))
        else: 
            model.add(_conv_activation_module())
    # Construct normalization layer, if it should be here.
    if norm_layer_position=='after':
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
    # ---------------------------------------------------------------------------
    # Check type and parameters of the pooling layer, and calculate output image size
    if pool_type is not None:
        if isinstance(pool_type, str) and 'global' not in pool_type.lower():
            temp_img_size = calc_image_size(img_size, kernel_size=pool_kernel_size, 
                                            stride=pool_stride, padding=pool_padding, dilation=1) # Keras pool has no support for dilation :(
            if min(temp_img_size) < min_image_dim:
                warnings.warn(
                "Output image ({}) of the pooling operation is smaller in one or more dimensions than min_image_dim={} for add_conv_block(). ".format(temp_img_size,min_image_dim)+ 
                "Using padding={} and stride=1 instead of padding={} and stride={}".format('same', pool_padding, pool_stride), UserWarning)
                pool_padding = 'same'
                pool_stride = 1
            _kwargs = {'pool_size':pool_kernel_size, 'strides':pool_stride, 'padding':pool_padding}
            if pool_params: _kwargs.update(pool_params)
            model.add(_pool_module(**_kwargs))
            img_size = calc_image_size(img_size, kernel_size=pool_kernel_size, stride=pool_stride, padding=pool_padding, dilation=1)
        else:
            if pool_params: model.add(_pool_module(**pool_params))
            else: model.add(_pool_module())
            if isinstance(pool_type, str) and 'global' in pool_type.lower():
                img_size = [1]*conv_dim
    # ---------------------------------------------------------------------------
    # Construct Dropout layer    
    if spatial_dropout: model.add(_spatial_dropout_module(spatial_dropout))
    if dropout: model.add(_dropout_module(dropout))
    # Store output image size as attribute    
    output_image = img_size
    d = {
        'model':model, 'output_image':output_image, 'out_channels':out_channels, 'norm_layer_position':norm_layer_position,
        'norm_layer_type':norm_layer_type, 'min_image_dim':min_image_dim, 'conv_padding':conv_padding, 'conv_stride':conv_stride, 'pool_padding':pool_padding,
        'pool_stride':pool_stride, 'conv_module':_conv_module, 'conv_activation_module':_conv_activation_module, 'pool_module':_pool_module,
        'dropout_module':_dropout_module, 'spatial_dropout_module':_spatial_dropout_module, 'norm_layer_module':_norm_layer_module
    }
    return d




if __name__ == '__main__':
    # Test the add_conv_block function
    model = tf.keras.models.Sequential()
    add_conv_block(model, out_channels=64, input_shape=[28, 28, 3], conv_dim=2, input_image=[28, 28], conv_kernel_size=3, conv_padding='valid',
    conv_stride=1, conv_dilation=1, conv_params=None, 
    conv_activation='relu', 
    conv_activation_params={'alpha':0.1}, norm_layer_position="before", 
    norm_layer_type="BatchNormalization", 
    norm_layer_params=None, pool_type="Max", pool_kernel_size=2, pool_padding='valid', pool_stride=1, pool_params=None, dropout=0.1, 
    spatial_dropout=None, min_image_dim=1, kernel_regularizer=None)
    model.summary()
    



if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *


# Variational Autoencoder with Keras
class Keras_VarAE(tf.keras.Model):
    def __init__(self, hparams, **kwargs):
        super(Keras_VarAE, self).__init__(**kwargs)

        self.num_features = hparams["num_features"]
        self.hidden_size = hparams["hidden_size"]
        self.latent_size = hparams["latent_size"]
        self.encoder_depth = hparams["encoder_depth"]
        self.encoder_activations = hparams["encoder_activations"]
        self.decoder_depth = hparams["decoder_depth"]
        self.decoder_activations = hparams["decoder_activations"]
        self.learning_rate = hparams["learning_rate"]
        self.learning_rate_decay_gamma = hparams.get("learning_rate_decay_gamma")
        self.optimizer = hparams["optimizer"]
        self.optimizer_params = hparams.get("optimizer_params")
        self.batch_size = hparams["batch_size"]
        self.epochs = hparams["epochs"]
        self.validation_tolerance_epochs = hparams.get("validation_tolerance_epochs")

        #self.inputlayer = Input(shape=(self.num_features,))
        self.encoder = tf.keras.models.Sequential(name="encoder")
        self.encoder.add(tf.keras.layers.Dense(
            self.hidden_size, input_shape=(self.num_features,), activation=self.encoder_activations,
            name="encoder_dense_1"))
        if self.encoder_depth > 1:
            for i in range(self.encoder_depth - 1):
                self.encoder.add(tf.keras.layers.Dense(self.hidden_size, activation=self.encoder_activations, 
                name="encoder_dense_" + str(i+2)))

        #self.encoder2mean = Sequential()
        #self.encoder2logvar = Sequential()
        #self.encoder2mean.add(Dense(self.latent_size, name="z_mean", input_shape=(self.hidden_size,)))
        #self.encoder2logvar.add(Dense(self.latent_size, name="z_logvar", input_shape=(self.hidden_size,)))
        self.encoder2mean = tf.keras.layers.Dense(self.latent_size, name="z_mean", input_shape=(self.hidden_size,))
        self.encoder2logvar = tf.keras.layers.Dense(self.latent_size, name="z_logvar", input_shape=(self.hidden_size,))
        
        self.decoder = tf.keras.models.Sequential(name="decoder")
        self.decoder.add(tf.keras.layers.Dense(
            self.hidden_size, input_shape=(self.latent_size,), activation=self.decoder_activations, 
            name="decoder_dense_1"))
        if self.decoder_depth > 1:
            for i in range(self.decoder_depth - 1):
                self.decoder.add(tf.keras.layers.Dense(
                    self.hidden_size, activation=self.decoder_activations, name="decoder_dense_" + str(i+2)))
        self.decoder.add(tf.keras.layers.Dense(self.num_features, activation="linear", name="decoder_dense_output"))

        #hid = self.encoder(self.inputlayer)
        #self.mu = self.encoder2mean(hid)
        #self.log_var = self.encoder2logvar(hid)
        #self.z = self.reparameterize(self.mu, self.log_var)
        #self.out_model = self.decoder(self.z)
        #self.out = Concatenate()([self.out_model, self.mu, self.log_var])
        #self.net = Model(self.inputlayer, self.out)
        

        self.total_loss_tracker = tf.keras.metrics.Mean(name="total_loss")
        self.reconstruction_loss_tracker = tf.keras.metrics.Mean(name="reconstruction_loss")
        self.kl_loss_tracker = tf.keras.metrics.Mean(name="kl_loss")

    @property
    def metrics(self):
        return [
            self.total_loss_tracker,
            self.reconstruction_loss_tracker,
            self.kl_loss_tracker,
        ]

    def train_step(self, data):
        with tf.GradientTape() as tape:
            hid = self.encoder(data)
            z_mean = self.encoder2mean(hid)
            z_log_var = self.encoder2logvar(hid)
            z = Sampling()([z_mean, z_log_var])
            reconstruction = self.decoder(z)
            reconstruction_loss = tf.reduce_mean(tf.keras.losses.mean_squared_error(data, reconstruction))
            kl_loss = -0.5 * (1 + z_log_var - tf.square(z_mean) - tf.exp(z_log_var))
            kl_loss = tf.reduce_mean(tf.reduce_sum(kl_loss, axis=1))
            total_loss = reconstruction_loss + kl_loss
        grads = tape.gradient(total_loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.trainable_weights))
        self.total_loss_tracker.update_state(total_loss)
        self.reconstruction_loss_tracker.update_state(reconstruction_loss)
        self.kl_loss_tracker.update_state(kl_loss)
        return {
            "loss": self.total_loss_tracker.result(),
            "reconstruction_loss": self.reconstruction_loss_tracker.result(),
            "kl_loss": self.kl_loss_tracker.result(),
        }

    
    def call(self, inputs):
        hid = self.encoder(inputs)
        z_mean = self.encoder2mean(hid)
        z_log_var = self.encoder2logvar(hid)
        z = Sampling()([z_mean, z_log_var])
        reconstruction = self.decoder(z)
        #return reconstruction, z_mean, z_log_var
        return reconstruction    






if __package__=="eznet_keras.models":
    from ..utils import *
else:
    import os, sys
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(parent_dir)
    from utils import *


class Dense_Block(tf.keras.layers.Layer):
    def __init__(self, input_shape:list=None, output_size:int=None, activation:str=None, activation_params:dict=None, dense_params:dict=None, norm_layer_type:str=None, norm_layer_position:str='before', 
                 norm_layer_params:dict=None, dropout:float=None, kernel_regularizer:tf.keras.regularizers.Regularizer=None):
        """Dense (fully connected) block containing one linear layer, followed optionally by a normalization layer, an activation function and a Dropout layer.

        ### Args:

            - `input_shape` (list|tuple): Shape of the input without the batch size.
            - `output_size` (int, optional): Number of output features. Defaults to None, in which case it will be input_size.
            - `activation` (str, optional): Activation. It can be an activation function name ("relu","sigmoid","tanh", etc.),
                an activation layer name ("ReLU", "LeakyReLU", "Softmax", etc.), or a custom Keras Layer class (not instance). Defaults to None.
            - `activation_params` (dict, optional): kwargs to pass to the activation function constructor. Defaults to None.
                Ignored if `activation` is a lower-case function name. By the way, the slope of the negative section in `LeakyReLU` is `alpha`.
            - `dense_params` (dict, optional): kwargs to pass to the dense layer constructor. Defaults to None.
                This will overwrite any other parameters such as regularizers, etc.
            - `norm_layer_type` (str, optional): Type of normalization layer. Defaults to None. Examples: 'BatchNormalization',
                'LayerNormalization', etc. It can also be a Keras Layer class (not instance).
            - `norm_layer_position` (str, optional): Position of norm layer relative to activation. Defaults to 'before'. Alternative is 'after'.
            - `norm_layer_params` (dict, optional): kwargs to pass to the norm layer constructor. Defaults to None.
            - `dropout` (float, optional): Dropout rate at the end. Defaults to None. Must be a float between 0 and 1.
            - `kernel_regularizer` (regularizer, optinal): Regularizer instance to be used for the kernel weights in the layers.
        
        
        **NOTE**: 
        Activation functions come from `tf.keras.activations` and need to be wrapped in a `tf.keras.layers.Activation` layer, and do not accept any
        kwargs. Activation layers come from `tf.keras.layers`, can accept kwargs in their constructors, and do not need to be wrapped in an
        `Activation` layer. Activation functions all have lower-case names, but activation layers are classes and have every-word-capitalized names.
        We will use this as a clue to recognize what the user wants, and perform accordingly.
        Most activation layers also have correpsonding activation functions (like `tf.keras.activations.relu` function and
        `tf.keras.layers.ReLU` layer). However, some activations are only available as functions (such as `tf.keras.activations.sigmoid`),
        and some are only available as layers (such as `tf.keras.layers.LeakyReLU`). Choose accordingly.
            
        ### Returns:
        
        A `tf.keras.layers.Layer` object.
        """
        super(Dense_Block, self).__init__()
        if output_size is None: 
            if input_shape:
                output_size = input_shape[-1]
            else:
                raise ValueError("Either input_shape or output_size must be provided.")
        self._input_shape = input_shape
        self._output_size = output_size
        self._activation = activation
        self._activation_params = activation_params
        self._norm_layer_type = norm_layer_type
        self._norm_layer_position = norm_layer_position
        self._norm_layer_params = norm_layer_params
        self._dense_params = dense_params
        self._dropout = dropout
        if activation is not None:
            if isinstance(activation, str):
                if activation.lower()==activation:
                    self._activation_module = getattr(tf.keras.activations, activation)
                else:
                    self._activation_module = getattr(tf.keras.layers, activation)
            else:
                self._activation_module = activation
        else:
            self._activation_module = None
        if norm_layer_type is not None:
            if isinstance(norm_layer_type, str):
                self._norm_layer_module = getattr(tf.keras.layers, norm_layer_type)
            else:
                self._norm_layer_module = norm_layer_type
        else:
            self._norm_layer_module = None
        self._dropout_module = tf.keras.layers.Dropout if dropout else None
        self._kernel_regularizer = kernel_regularizer
        self.net = tf.keras.models.Sequential()
        _kwargs = {
            'units': output_size
        }
        if input_shape:
            _kwargs.update({'input_shape':input_shape})
        if kernel_regularizer is not None:
            _kwargs.update({'kernel_regularizer':kernel_regularizer})
        if dense_params is not None:
            _kwargs.update(dense_params)
        self.net.add(tf.keras.layers.Dense(**_kwargs))
        if norm_layer_type and norm_layer_position=='before': 
            if norm_layer_params: self.net.add(self._norm_layer_module(**norm_layer_params))
            else: self.net.add(self._norm_layer_module())
        if activation:
            if isinstance(activation, str) and activation.lower()==activation: 
                self.net.add(tf.keras.layers.Activation(self._activation_module))
            else:
                if activation_params: self.net.add(self._activation_module(**activation_params))
                else: self.net.add(self._activation_module())
        if norm_layer_type and norm_layer_position=='after': 
            if norm_layer_params: self.net.add(self._norm_layer_module(**norm_layer_params))
            else: self.net.add(self._norm_layer_module())
        if dropout: self.net.add(self._dropout_module(dropout))
    
    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)
    
    def get_config(self):
        config = super(Dense_Block, self).get_config()
        hparams = {
            'input_shape':self._input_shape,
            'output_size':self._output_size,
            'activation':self._activation,
            'activation_params':self._activation_params,
            'norm_layer_type':self._norm_layer_type,
            'norm_layer_position':self._norm_layer_position,
            'norm_layer_params':self._norm_layer_params,
            'dropout':self._dropout,
            'kernel_regularizer':self._kernel_regularizer,
            'dense_params':self._dense_params
        }
        config['hparams'] = hparams
        return config
    
    @classmethod
    def from_config(cls, config):
        return cls(**config['hparams'])
    
    def summary(self):
        return self.net.summary()
        



def add_dense_block(model:tf.keras.models.Sequential, output_size:int, input_shape:list=None,
                    activation:str=None, activation_params:dict=None, dense_params:dict=None,
                    norm_layer_type:str=None, norm_layer_position:str='before', norm_layer_params:dict=None, dropout:float=None, 
                    kernel_regularizer:tf.keras.regularizers.Regularizer=None):
    """
    Add a Dense (fully connected) block containing one linear layer, followed optionally by a normalization layer, an activation function and a Dropout layer, 
    to a `tf.keras.models.Sequential` instance.

    ### Args:
        
        - `input_shape` (list|tuple, optional): Shape of the input disregarding the batch size.
        - `output_size` (int, optional): Number of output features. Defaults to None, in which case it will be input_size.
        - `activation` (str, optional): Activation. It can be an activation function name ("relu","sigmoid","tanh", etc.),
            an activation layer name ("ReLU", "LeakyReLU",
            "Softmax", etc.), or a custom Keras Layer class (not instance). Defaults to None.
        - `activation_params` (dict, optional): kwargs to pass to the activation function constructor. Defaults to None.
            Ignored if `activation` is a lower-case function name. By the way, the slope of the negative section in `LeakyReLU` is `alpha`.
        - `dense_params` (dict, optional): kwargs to pass to the dense layer constructor. Defaults to None.
            This will overwrite any other parameters such as regularizers, etc.
        - `norm_layer_type` (str, optional): Type of normalization layer. Defaults to None. Examples: 'BatchNormalization',
            'LayerNormalization', etc. It can also be a Keras Layer class (not instance).
        - `norm_layer_position` (str, optional): Position of norm layer relative to activation. Defaults to 'before'. Alternative is 'after'.
        - `norm_layer_params` (dict, optional): kwargs to pass to the norm layer constructor. Defaults to None.
        - `dropout` (float, optional): Dropout rate at the end. Defaults to None. Must be a float between 0 and 1.
        - `kernel_regularizer` (regularizer, optinal): Regularizer to be used for the kernel weights in the layers.
            
    ### Returns:
        
        Nothing. It modifies the `model` argument passed to it.
        
    """
    
    _dropout_module = tf.keras.layers.Dropout if dropout else None
    
    if activation is not None:
        if isinstance(activation, str):
            if activation.lower()==activation:
                _activation_module = getattr(tf.keras.activations, activation)
            else:
                _activation_module = getattr(tf.keras.layers, activation)
        else:
            _activation_module = activation
    else:
        _activation_module = None
        
    if norm_layer_type is not None:
        if isinstance(norm_layer_type, str):
            _norm_layer_module = getattr(tf.keras.layers, norm_layer_type)
        else:
            _norm_layer_module = norm_layer_type
    else:
        _norm_layer_module = None
    
    
    _kwargs = {
        'units': output_size
    }
    if input_shape is not None:
        _kwargs.update({'input_shape':input_shape})
    if kernel_regularizer is not None:
        _kwargs.update({'kernel_regularizer':kernel_regularizer})
    if dense_params is not None:
        _kwargs.update(dense_params)
    model.add(tf.keras.layers.Dense(**_kwargs))
        
    if norm_layer_type and norm_layer_position=='before': 
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
        
    if activation: 
        if isinstance(activation, str) and activation.lower()==activation: 
            model.add(tf.keras.layers.Activation(_activation_module))
        else:
            if activation_params: model.add(_activation_module(**activation_params))
            else: model.add(_activation_module())
            
    if norm_layer_type and norm_layer_position=='after': 
        if norm_layer_params: model.add(_norm_layer_module(**norm_layer_params))
        else: model.add(_norm_layer_module())
        
    if dropout: model.add(_dropout_module(dropout))
    
    
    

if __name__=='__main__':
    
    # Test Dense_Block class
    print("Testing Dense_Block class...")
    block = Dense_Block(input_shape=[5], output_size=20, activation="LeakyReLU", activation_params={'alpha':0.1}, norm_layer_type="BatchNormalization", 
                        norm_layer_position='before', norm_layer_params=None, dropout=0.1, kernel_regularizer=None)
    block.summary()
    y = block(tf.random.normal([32,5]))
    print("Shape of output tensor:", y.shape)
    
    print("\n")
    print("Testing add_dense_block function...")
    model = tf.keras.models.Sequential()
    add_dense_block(model, output_size=20, input_shape=[10], activation="LeakyReLU", activation_params={'alpha':0.1}, 
                    norm_layer_type="BatchNormalization", norm_layer_position='before', norm_layer_params=None, dropout=0.5, 
                    kernel_regularizer=None)
    model.summary()
    y = model(tf.random.normal([32,10]))
    print("Shape of output tensor:", y.shape)
    


if __package__=="eznet_keras.models":
    from .keras_smart_module import *
    from .conv_block import *
    from .dense_block import *
else:
    import sys, os
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
    from keras_smart_module import *
    from conv_block import *
    from dense_block import *

class Conv_Network(KerasSmartModel):
    
    sample_hparams = {
        "model_name": "Conv_Network",
        # I/O shapes (without the batch dimension)
        "input_shape": [28, 28, 3],
        "output_shape": [10],
        # Convolution blocks
        "num_conv_blocks": 2,
        "conv_dim": 2,
        "conv_params": None,
        "conv_channels": "auto",
        "conv_kernel_size": 3,
        "conv_padding": "valid",
        "conv_stride": 1,
        "conv_dilation": 1,
        "conv_activation": "relu",
        "conv_activation_params": None,
        "conv_norm_layer_type": "BatchNormalization",
        "conv_norm_layer_position": "before",
        "conv_norm_layer_params": None,
        "conv_dropout": 0.1,
        "conv_spatial_dropout": None,
        "pool_type": "Max",
        "pool_kernel_size": 2,
        "pool_padding": 'valid',
        "pool_stride": 1,
        "pool_params": None,
        "min_image_size": 4,
        "flatten_after_conv": True,
        # Fully connected blocks
        "include_dense_layers": True,
        "dense_width": "auto",
        "dense_depth": 2,
        "dense_params": None,
        "dense_activation": "relu",
        "dense_activation_params": None,
        "dense_norm_layer_type": "BatchNormalization",
        "dense_norm_layer_position": "before",
        "dense_norm_layer_params": None,
        "dense_dropout": 0.1,
        "include_output_layer": True,
        "output_dense_params": None,
        "output_activation": "softmax",
        "output_activation_params": None,
        # Training procedure
        "l2_reg": 0.0001,
        "l1_reg": None,
        "batch_size": 32,
        "epochs": 2,
        "validation_data": [0.05,'testset'],
        "early_stopping_patience_epochs": 5,
        "learning_rate": 0.01,
        "exponential_decay_rate": 0.9,
        "loss_function": "categorical_crossentropy",
        "loss_function_params": None,
        "optimizer": "Adam",
        "optimizer_params": None,
        'metrics':['accuracy'],
        'metrics_params':None,
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6,
        'other_callbacks': None,
        'custom_schedule': None,
    }
    
    
    def __init__(self, hparams:dict=None):
        """Standard Convolutional Neural Network, containing convolutional blocks followed by fully-connected blocks. It supports 1D, 2D, and 3D
        convolutions, and can be used for image classification, timeseries classification, video classification, and so forth.
        The module can easily be trained and evaluated using its own methods, because it inherits from `KerasSmartModel`.
        The architecture consists of conv blocks followed by dense (fully connected) blocks. Each conv block is assumed to contain a convolution
        layer, followed optionally be a normalization layer, an activation, pooling, and finally a dropout or spatial dropout layer.
        Each dense block is assumed to contain a dense layer, followed optionally be a normalziation layer, an activation, and a dropout layer.
        Convolution blocks are mandatory, but Dense layers and an output layer are optional.
        Using this class, any kind of serial CNN architecture can be built. However, no parallelism or skip connections are supported in the
        architecture.

        ### Usage

        `model = Conv_Network(hparams)` where `hparams` is dictionary of hyperparameters containing the following keys. 
        
        - Inspect the `sample_hparams` class attribute for a template of the hyperparameters dictionary.
        - Many keys, especially those that have "(list of)" in their description, can be either scalar items or lists of items.
          If they are scalar, they will be broadcasted to all blocks. If they are lists, they must have the same length as the number (depth) of
          convolutional blocks, i.e. `num_conv_blocks` hyperparameter.
        - Every block by default is assumed to have a convolution operation, a normalization layer, an activation, some form of pooling, and finally
          a dropout or spatial dropout. 
          For every convolutional block, the convolution operation is mandatory but the rest are optional. This way, any kind of CNN with any
          (fully serial) architecture can be built. If the key to some hyperparameter is a list, it must have a length equal to the number of
          convolutional blocks, that is, the `num_conv_blocks` hyperparameter. 
          There must be one item per convolutional block. `None` is typically put for every block that doesn't have that layer or that hyperparameter
          is inapplicable to that block.
          Also note that we assume every convolutional block has a spatial dropout layer followed by a normal dropout layer, but this is not common
          in reality, which means for blocks that have spatial dropout the normal dropout rate should be `None`, and vice versa.
        - The Training Procedure section of the hyperparameters are optional, and will only be used by the parent class `KerasSmartModel` if the
          model is trained using its methods. Otherwise, it is not necessary at all.
        - Activation functions come from `tf.keras.activations` and need to be wrapped in a `tf.keras.layers.Activation` layer, and do not accept any
          kwargs. Activation layers come from `tf.keras.layers`, can accept kwargs in their constructors, and do not need to be wrapped in an
          `Activation` layer. Activation functions all have lower-case names, but activation layers are classes and have every-word-capitalized
          names. We will use this as a clue to recognize what the user wants, and perform accordingly.
          Most activation layers also have correpsonding activation functions (like `tf.keras.activations.relu` function and
          `tf.keras.layers.ReLU` layer). However, some activations are only available as functions (such as `tf.keras.activations.sigmoid`),
          and some are only available as layers (such as `tf.keras.layers.LeakyReLU`). Choose accordingly.

        #### I/O shapes
        
        - `input_shape` (list): Input shape *WITHOUT* the batch dimension. For instance, for 2D images, input should be [N, H, W, C],
          therefore `input_shape` should be [H, W, C].
        - `output_shape` (int): Output shape *WITHOUT* the batch dimension. For instance, for K-class classification, model outputs can be [N, K],
          so `output_shape` should be [K].
            
        #### Convolution blocks
        
        - `num_conv_blocks` (int): Number of convolutional blocks. Every block contains a convolutional layer, and
            optionally a normalization layer, an activation layer, a pooling layer, and finally a dropout or spatial dropout layer.
        - `conv_dim` (int): Dimensionality of the convolution. 1, 2, or 3.
        - `conv_params` (dict): (list of) kwargs dict to pass to the convolution constructor in each block. Defaults to None.
        - `conv_channels` (int|list|str): (list of) Number of filters of the convolution layer in each conv block. If `"auto"`, it will start
            with the input channels, and double with every block, in powers of two. If `list`, it should be a list
            of channels for each conv block. If `int`, it will be the same for all conv blocks. Default is `"auto"`.
        - `conv_kernel_size` (int|list): (list of) Kernel size of the convolution layers. Should be a list of integers,
            a list of tuples of integers (for conv2d or conv3d), or an integer. If it is a list, it MUST have the same 
            length as `num_conv_blocks`. If it is an integer, it will be the same for all conv blocks. Defaults to `3`.
        - `conv_padding` (int|str|list): (list of) Paddings of convolution layers. Format is as `conv_kernel_size`. Defaults to `"valid"`.
        - `conv_stride` (int|list): (list of) Strides of convolution layers. Format is as `conv_kernel_size`. Defaults to `1`.
        - `conv_dilation` (int|list): (list of) Dilations of convolution layers. Format is as `conv_kernel_size`. Defaults to `1`.
        - `conv_activation` (str|list): (list of) activations of the convolution blocks. Defaults to None.
            For each block, this entry can be an activation function ("relu", "sigmoid", "tanh", etc.), an activation layer 
            ("ReLU", "LeakyReLU", "Softmax", etc.), or a custom Layer class (not instance).
            `None` will assume no activation for the convolution layer.
        - `conv_activation_params` (dict|list): (list of) dicts for the convolution activation constructors. Defaults to None.
            This will be ignored if lower-case activation
            function names are provided. This is because in Keras, activation layers are classes with constructors, but activation functions are
            just functions, and have lower-case names as mentioned earlier. By the way, the slope of the negative section in `LeakyReLU` is `alpha`.
        - `conv_norm_layer_type` (str|list): (list of) types of normalization layers to use in the conv blocks. Examples: 'BatchNormalization',
            'LayerNormalization', etc. Defaults to None. Instead of a string, it can also be a custom Keras Layer class (not instance).
        - `conv_norm_layer_position` ("before"|"after"|None|list): (list of) positions of the normalization layers in the 
            convolutional blocks relative to the activation functions. Defaults to "before". If it is a list, it should be a list of strings of the same length as `num_conv_blocks`
        - `conv_norm_layer_params` (dict|list): (list of) kwargs dicts for the convolution normalization layers' constructors. Defaults to None.    
        - `conv_dropout` (float|list): (list of) Dropout rates of the convolution blocks. Defaults to None.
        - `conv_spatial_dropout` (float|list): (list of) Spatial dropout rates of the convolution blocks. Defaults to None.
        - `pool_type` (str|list): (list of) types of pooling layers. "Max", "Avg", "GlobalMax", and "GlobalAvg" are acceptable.
            Defaults to None, in which case there will be no pooling layer. Instead of the strings mentioned above, a custom Keras Layer class
            (not instance) can also be provided. In this case, one should note that the pooling
            hyperparameters below will be ignored, and the custom pooling layer will be constructed with the `pool_params` kwargs only.
            Also, in this case, image dimensions will not be calculated for this pooling layer because it is unknown since it is a custom layer.
        - `pool_kernel_size` (int|list): (list of) kernel sizes of the pooling layers, with similar format to `conv_kernel_size`.
            Again, it can be a list of integers, where every integer will be broadcasted to all dimensions, a list of tuples of integers for
            2D or 3D images, or an integer to be boradcasted across dimensions and conv blocks.
        - `pool_padding` (str|list): (list of) paddings of the pooling layers.
        - `pool_stride` (int|list): (list of) strides of the pooling layers.
        - `pool_params` (dict|list): (list of) kwargs dicts for the pooling layers' constructors.
        - `min_image_size` (int): Minimum size of the image to be reduced to in convolutions and poolings.
            After this point, the padding and striding will be chosen such that image size does not decrease further. Defaults to 1.
        - `flatten_before_dense` (bool): Whether to flatten the output of the convolutional blocks before the dense blocks.
            If it is False, all dense layers (if any) will be applied to all pixels, acting on channels. If this argument is not provided, the default will depend on whether there are dense layers after conv layers,
            in which case `dense_depth` would be greater than 0 and the default would be True, or not, in which case the default would be
            False since the absence of dense layers negates any reason for flattening. Defaults to None.
            
        #### Dense blocks
        
        - `dense_width` ("auto"|int|list): Width of the hidden layers of the Dense network. "auto", a number (for all of them) or a list
            holding width of each hidden layer.
            If "auto", it will start with the output size of the Flatten() layer, halving at every Dense block.
        - `dense_depth` (int): Depth (number of hidden layers) of the Dense network. `0` will mean no hidden layers,
            meaning Flatten will be directly followed by the output layer.
        - `dense_params` (dict): (list of) kwargs dict to pass to the dense layer constructor in each block. Defaults to None.
        - `dense_activation` (str|list): (list of) activation function for hidden layers of the Dense network.
            These can be activation functions ("relu", "sigmoid", "tanh", etc.),
            activation layers ("ReLU", "LeakyReLU", "Softmax", etc.), or custom Layer classes (not instances). Defaults to None,
            in which case no activation function will be used.
        - `dense_activation_params` (dict|list): (list of) dicts for the dense activation functions' constructors.
            Ignored if lower-case activation functions are provided.
            By the way, the slope of the negative section in `LeakyReLU` is `alpha`.
        - `dense_norm_layer_type` (str|list): (list of) types of normalization layers to use in the dense blocks. Examples:
            'BatchNormalization', 'LayerNormalization', etc.
            Defaults to None, in which case no normalization layer will be used. Instead of strings, custom Keras Layer classes (not instances)
            can also be provided.
        - `dense_norm_layer_position` ("before"|"after"|list): (list of) positions of the normalization layers in the dense blocks relative to the
            activation functions. Defaults to "before". If it is a list, it should be a list of strings of the same length as `dense_depth`.
        - `dense_norm_layer_params` (dict|list): (list of) kwargs dict for the dense normalization layers' constructors.
        - `dense_dropout` (float|list): (list of) Dropout rates (if any) for the hidden layers of the Dense network.
        - `include_output_layer` (bool): Whether to include an output layer. Defaults to True.
        - `output_dense_params` (dict): kwargs dict for the output layer's Dense constructor. Defaults to None.
        - `output_activation` (str): Activation (with the same format as dense activation) for the output layer, if any.
            **NOTE** Depending on the loss function, you may not need an activation function.
            For classification problems, you may want to choose "sigmoid" or "softmax".
            That being said, you usually don't need to specify an activation for the output layer at all, if e.g. 'from_logits' is used.
            For regression problems, no activation is needed. It is by default linear, unless you want to manually specify an activation.
        - `output_activation_params` (dict): Dictionary of parameters for the output activation function's constructor.
            Ignored if a lower-case function name is provided.
        
        #### Training procedure
        
        - `batch_size` (int): Minibatch size, the expected input size of the network.
        - `learning_rate` (float): Initial learning rate of training.
        - `exponential_decay_rate` (float): Exponential decay rate gamma for learning rate, if any.
        - `optimizer` (str): Optimizer. Examples: 'Adam', 'SGD', 'RMSprop', etc. It can be the name of any Keras optimizer class.
            This can also be a custom optimizer class (not instance), in which case `optimizer_params` can be specified for its constructor kwargs.
        - `optimizer_params` (dict): Additional parameters of the optimizer constructor, if any.
        - `epochs` (int): Maximum number of epochs for training.
        - `early_stopping_patience_epochs` (int): Epochs to tolerate unimproved (val) loss, before early stopping.
        - `l2_reg` (float): L2 regularization parameter.
        - `l1_reg` (float): L1 regularization parameter.
        - `loss_function` (str): Loss function. It can be a lower-case name such as "mse", "binary_crossentropy", "categorical_crossentropy", etc.,
            the name of a `tf.keras.losses` class such as `BinaryCrossentropy`, `CategoricalCrossentropy`, `MeanSquaredError', etc., or a valid loss
            class (not instance).
        - `loss_function_params` (dict): Additional kwargs parameters of the loss function constructor, if any. Ignored if the loss function is a
            lower-case name string.
        - `metrics` (list): list of metrics for Keras compilation. Each member of the list can be a lower-case metric name such as
            "mse" or "accuracy", the name of a `tf.keras.metrics` class such as `Accuracy`, `MeanSquaredError`, etc., or a valid metric class
            (not instance).
        - `metrics_params` (list): (list of) additional kwargs parameters of the metrics constructors, if any.
            Ignored for every metric that is a lower-case name string.
            If this entry is a single dicitonary rather than a list, it will be broadcast to all metrics in the metrics list.
        - `validation_data` (tuple): Validation data, if any. It should be a tuple of `(portion, from_dataset)`.
            For instance, `[0.05, 'testset']` means 5% of the testset will be used for validation.
            The second element of the tuple can only be `'trainset'` and `'testset'`. The first element must be a float between 0 and 1. 
            If the second element is not specified, testset will be used by default.
        - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
        - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
        - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping.
            Defaults to 'min' for any error. 'max' is for accuracy, etc.
        - `early_stopping_value` (float): Value of the monitor at which point training will stop because the critical value has been reached.
        - `other_callbacks` (list): List of other callbacks to be used during training, if any. Defaults to None.
        - `custom_schedule` (schedule): Custom learning rate schedule inheriting from `tf.keras.optimizers.schedules.LearningRateSchedule`.
            Defaults to None.
        
        
        ### Returns
        
        - Returns a `tf.keras.models.Model` object that can be trained and used accordingly.
        - Run `net.summary()` afterwards to see what you have inside the network.
        - A `KerasSmartModel` object is returned. This module has its own functions for training, evaluation, etc.
        """
        super(Conv_Network, self).__init__(hparams)
        if not hparams: hparams = self.sample_hparams
        # Input and output shapes
        self._model_name = hparams["model_name"] if "model_name" in hparams else "Conv_Network"
        self._input_shape = hparams["input_shape"]
        self._output_shape = hparams["output_shape"]
        self._N = int(hparams["batch_size"]) if "batch_size" in hparams else 32
        self.batch_input_shape = list(self._input_shape).copy()
        self.batch_input_shape.insert(0, self._N)
        self.batch_output_shape = list(self._output_shape).copy()
        self.batch_output_shape.insert(0, self._N)
        self.size_list = [self._input_shape]
    
        # Initialiaing sequential network
        self.net = tf.keras.models.Sequential(name=self._model_name)
        
        # Convolutional layers hyperparameters
        self._num_conv_blocks = hparams["num_conv_blocks"]
        self._conv_dim = hparams["conv_dim"]
        self._conv_params = hparams.get("conv_params")    
        self._conv_channels = hparams.get("conv_channels") if hparams.get("conv_channels") else "auto"
        self._conv_kernel_size = hparams.get("conv_kernel_size") if hparams.get("conv_kernel_size") else 3
        self._conv_padding = hparams["conv_padding"] if hparams.get("conv_padding") else "valid"
        self._conv_stride = hparams["conv_stride"] if hparams.get("conv_stride") else 1
        self._conv_dilation = hparams["conv_dilation"] if hparams.get("conv_dilation") else 1
        self._conv_activation = hparams.get("conv_activation")
        self._conv_activation_params = hparams.get("conv_activation_params")
        self._conv_norm_layer_type = hparams.get("conv_norm_layer_type")
        self._conv_norm_layer_position = hparams.get("conv_norm_layer_position")
        self._conv_norm_layer_params = hparams.get("conv_norm_layer_params")
        self._conv_dropout = hparams.get("conv_dropout")
        self._conv_spatial_dropout = hparams.get("conv_spatial_dropout")
        self._pool_type = hparams.get("pool_type")
        self._pool_kernel_size = hparams["pool_kernel_size"] if hparams.get("pool_kernel_size") else 2
        self._pool_padding = hparams["pool_padding"] if hparams.get("pool_padding") else 'valid'
        self._pool_stride = hparams["pool_stride"] if hparams.get("pool_stride") else 1
        self._pool_params = hparams.get("pool_params")
        self._min_image_size = hparams["min_image_size"] if hparams.get("min_image_size") else 1        
        
        # Generate lists of hyperparameters for conv/pool layers
        self._conv_channels_vec = self._gen_hparam_vec_for_conv(self._conv_channels, "conv_channels", 
            check_auto=True, init_for_auto=self._input_shape[-1], powers_of_two_if_auto=True, direction_if_auto="up")
        self._conv_kernel_size_vec = self._gen_hparam_vec_for_conv(self._conv_kernel_size, "conv_kernel_size")
        self._pool_kernel_size_vec = self._gen_hparam_vec_for_conv(self._pool_kernel_size, 'pool_kernel_size')
        self._conv_padding_vec = self._gen_hparam_vec_for_conv(self._conv_padding, 'conv_padding')
        self._pool_padding_vec = self._gen_hparam_vec_for_conv(self._pool_padding, 'pool_padding')
        self._conv_stride_vec = self._gen_hparam_vec_for_conv(self._conv_stride, 'conv_stride')
        self._pool_stride_vec = self._gen_hparam_vec_for_conv(self._pool_stride, 'pool_stride')
        self._conv_dilation_vec = self._gen_hparam_vec_for_conv(self._conv_dilation, 'conv_dilation')
        self._conv_activation_vec = self._gen_hparam_vec_for_conv(self._conv_activation, 'conv_activation')
        self._conv_activation_params_vec = self._gen_hparam_vec_for_conv(self._conv_activation_params, 'conv_activation_params')
        self._pool_type_vec = self._gen_hparam_vec_for_conv(self._pool_type, 'pool_type')
        self._pool_params_vec = self._gen_hparam_vec_for_conv(self._pool_params, 'pool_params')
        self._conv_params_vec = self._gen_hparam_vec_for_conv(self._conv_params, 'conv_params')
        self._conv_norm_layer_type_vec = self._gen_hparam_vec_for_conv(self._conv_norm_layer_type, 'conv_norm_layer_type')
        self._conv_norm_layer_params_vec = self._gen_hparam_vec_for_conv(self._conv_norm_layer_params, 'conv_norm_layer_params')
        self._conv_norm_layer_position_vec = self._gen_hparam_vec_for_conv(self._conv_norm_layer_position, 'conv_norm_layer_position')
        self._conv_dropout_vec = self._gen_hparam_vec_for_conv(self._conv_dropout, 'conv_dropout')
        self._conv_spatial_dropout_vec = self._gen_hparam_vec_for_conv(self._conv_spatial_dropout, 'conv_spatial_dropout')
        
        # Constructing the encoder (convolutional blocks)
        # print("input_shape: ", self._input_shape)
        # in_channels = self._input_shape[-1]
        input_image = list(self._input_shape[:-1])
        for i in range(self._num_conv_blocks):
            out_channels = self._conv_channels_vec[i]
            # print("in_channels: ", in_channels)
            # print("out_channels: ", out_channels)
            # print("input_image: ", input_image)
            _kwargs = {
                'model':self.net, 
                'out_channels':out_channels, 
                'conv_dim':self._conv_dim, 
                'input_image':input_image, 
                'conv_kernel_size':self._conv_kernel_size_vec[i], 
                'conv_padding':self._conv_padding_vec[i],
                'conv_stride':self._conv_stride_vec[i], 
                'conv_dilation':self._conv_dilation_vec[i], 
                'conv_params':self._conv_params_vec[i], 
                'conv_activation':self._conv_activation_vec[i], 
                'conv_activation_params':self._conv_activation_params_vec[i], 
                'norm_layer_position':self._conv_norm_layer_position_vec[i], 
                'norm_layer_type':self._conv_norm_layer_type_vec[i], 
                'norm_layer_params':self._conv_norm_layer_params_vec[i], 
                'pool_type':self._pool_type_vec[i], 
                'pool_kernel_size':self._pool_kernel_size_vec[i], 
                'pool_padding':self._pool_padding_vec[i], 
                'pool_stride':self._pool_stride_vec[i], 
                'pool_params':self._pool_params_vec[i], 
                'dropout':self._conv_dropout_vec[i],
                'spatial_dropout':self._conv_spatial_dropout_vec[i], 
                'min_image_dim':self._min_image_size,
                'kernel_regularizer':self.make_regularizer()
            }
            if i==0:
                _kwargs.update({'input_shape':self._input_shape})
            d = add_conv_block(**_kwargs)
            # self.net = d['model']
            output_image = d['output_image']
            self.size_list.append(output_image+[out_channels])
            # in_channels = out_channels
            input_image = output_image
            
        # Including Dense or Output layers
        self._dense_depth = hparams.get("dense_depth")
        self._include_dense_layers = hparams["include_dense_layers"] if "include_dense_layers" in hparams else \
            (self._dense_depth is not None and self._dense_depth > 0)
        self._include_output_layer = hparams["include_output_layer"] if "include_output_layer" in hparams else True
        
        # Flattening
        self._flatten_after_conv = hparams["flatten_after_conv"] if "flatten_after_conv" in hparams else (len(self._output_shape)==1)
        if self._flatten_after_conv:
            self.net.add(tf.keras.layers.Flatten())
            self._dense_input_size = np.prod(output_image) * out_channels
        else:
            self._dense_input_size = out_channels
        self.size_list.append([self._dense_input_size])
        
        # Construct dense layers
        if self._include_dense_layers:
            
            # Dense layers hyperparameters
            self._dense_width = hparams.get("dense_width")
            self._dense_params = hparams.get("dense_params")
            self._dense_activation = hparams.get("dense_activation")
            self._dense_activation_params = hparams.get("dense_activation_params")
            self._dense_norm_layer_type = hparams.get("dense_norm_layer_type")
            self._dense_norm_layer_params = hparams.get("dense_norm_layer_params")
            self._dense_norm_layer_position = hparams.get("dense_norm_layer_position")
            self._dense_dropout = hparams.get("dense_dropout")
            
            # Generate lists of hyperparameters for the dense layers
            self._dense_width_vec = self._gen_hparam_vec_for_dense(self._dense_width, 'dense_width',
                check_auto=True, init_for_auto=self._dense_input_size, powers_of_two_if_auto=True, direction_if_auto="down")
            self._dense_params_vec = self._gen_hparam_vec_for_dense(self._dense_params, 'dense_params')
            self._dense_activation_vec = self._gen_hparam_vec_for_dense(self._dense_activation, 'dense_activation')
            self._dense_activation_params_vec = self._gen_hparam_vec_for_dense(self._dense_activation_params, 'dense_activation_params')
            self._dense_norm_layer_type_vec = self._gen_hparam_vec_for_dense(self._dense_norm_layer_type, 'dense_norm_layer_type')
            self._dense_norm_layer_params_vec = self._gen_hparam_vec_for_dense(self._dense_norm_layer_params, 'dense_norm_layer_params')
            self._dense_norm_layer_position_vec = self._gen_hparam_vec_for_dense(self._dense_norm_layer_position, 'dense_norm_layer_position')
            self._dense_dropout_vec = self._gen_hparam_vec_for_dense(self._dense_dropout, 'dense_dropout')
            
            # Construct the dense layers
            # in_size = self._dense_input_size
            if self._dense_depth > 0:
                for i in range(self._dense_depth):
                    out_size = self._dense_width_vec[i]
                    _kwargs = {
                        'model':self.net,
                        'output_size':self._dense_width_vec[i],
                        'activation':self._dense_activation_vec[i],
                        'activation_params':self._dense_activation_params_vec[i],
                        'norm_layer_type':self._dense_norm_layer_type_vec[i],
                        'norm_layer_position':self._dense_norm_layer_position_vec[i],
                        'norm_layer_params':self._dense_norm_layer_params_vec[i],
                        'dropout': self._dense_dropout_vec[i],
                        'kernel_regularizer':self.make_regularizer(),
                        'dense_params':self._dense_params_vec[i]
                    }
                    add_dense_block(**_kwargs)
                    # in_size = out_size
                    self.size_list.append([out_size])
        
        # Construct output layer
        if self._include_output_layer:
            # Output hyperparameters
            self._output_dense_params = hparams.get("output_dense_params")
            self._output_activation = hparams.get("output_activation")
            self._output_activation_params = hparams.get("output_activation_params")
        
            # Output layer
            _kwargs = {'units':self._output_shape[-1], 'kernel_regularizer':self.make_regularizer()}
            if self._output_dense_params:
                _kwargs.update(self._output_dense_params)
            self.net.add(tf.keras.layers.Dense(**_kwargs))
            if self._output_activation:
                if isinstance(self._output_activation, str):
                    if self._output_activation.lower() == self._output_activation:
                        self.net.add(tf.keras.layers.Activation(self._output_activation))
                    elif self._output_activation_params:
                        self.net.add(getattr(tf.keras.layers, self._output_activation)(**self._output_activation_params))
                    else:
                        self.net.add(getattr(tf.keras.layers, self._output_activation)())
                elif self._output_activation_params:
                    self.net.add(self._output_activation(**self._output_activation_params))
                else:
                    self.net.add(self._output_activation())

    
    def _gen_hparam_vec_for_conv(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._num_conv_blocks, hparam_name=hparam_name, count_if_not_list_name='num_conv_blocks', **kwargs)
    
    def _gen_hparam_vec_for_dense(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._dense_depth, hparam_name=hparam_name, count_if_not_list_name='dense_depth', **kwargs)
    
    def call(self, inputs, **kwargs):
        return self.net(inputs, **kwargs)





if __name__ == '__main__':
    # Test Conv_Network
    print("Minimal example with only convolution layers and no dense or output layers:")
    sample_hparams = {
        "model_name": "Conv_Network",
        # I/O shapes (without the batch dimension)
        "input_shape": [28, 28, 3],
        "output_shape": [10],
        # Convolution blocks
        "num_conv_blocks": 2,
        "conv_dim": 2,
        "conv_params": None,
        "conv_channels": "auto",
        "conv_kernel_size": 3,
        "conv_padding": "valid",
        "conv_stride": 1,
        "conv_dilation": 1,
        "conv_activation": "LeakyReLU",
        "conv_activation_params": {'alpha':0.1},
        "conv_norm_layer_type": "BatchNormalization",
        "conv_norm_layer_position": ["before", "after"],
        "conv_norm_layer_params": None,
        "conv_dropout": [0.1, None],
        "conv_spatial_dropout": None,
        "pool_type": "Max",
        "pool_kernel_size": 2,
        "pool_padding": 'valid',
        "pool_stride": 1,
        "pool_params": None,
        "min_image_size": 4
    }
    model = Conv_Network(sample_hparams)
    model.summary()
    print("Trying with sample data ...")
    y = model(tf.random.normal([32, 28, 28, 3]))
    print("Output shape:", y.shape)
    
    
    print("")
    print("Testing full example with convolution layers, dense layers and output layer:")
    test_keras_model_class(Conv_Network)
    